import json
import os
import shutil
import tempfile
import unittest
from contextlib import redirect_stdout
from io import StringIO
from unittest.mock import MagicMock, patch

from netorca_sdk.exceptions import (
    NetorcaAPIError,
    NetorcaAuthenticationError,
    NetorcaException,
    NetorcaGatewayError,
    NetorcaNotFoundError,
    NetorcaServerUnavailableError,
)
from netorca_sdk.serviceowner import ServiceOwnerSubmission


class TestServiceOwnerSubmission(unittest.TestCase):
    def setUp(self):
        self.netorca_api_key = "api_key"
        self.service_owner_submission = ServiceOwnerSubmission(self.netorca_api_key)

        # Create a temporary directory and set up test files
        self.test_dir = tempfile.mkdtemp()
        os.makedirs(os.path.join(self.test_dir, ".netorca"))

        with open(os.path.join(self.test_dir, ".netorca", "config.json"), "w") as f:
            f.write('{"netorca_global": {"base_url": "https://example.com"}}')

        with open(os.path.join(self.test_dir, ".netorca", "service1.json"), "w") as f:
            f.write('{"key": "value"}')

        with open(os.path.join(self.test_dir, ".netorca", "service1.md"), "w") as f:
            f.write("# Service 1 Documentation")

        # Mock the NetorcaAuth.post method and requests.post method
        self.mock_post_patcher = patch("netorca_sdk.auth.NetorcaAuth.post")
        self.mock_post = self.mock_post_patcher.start()
        self.mock_requests_post_patcher = patch("requests.post")
        self.mock_requests_post = self.mock_requests_post_patcher.start()

    def tearDown(self):
        # Remove temporary directory and its contents
        shutil.rmtree(self.test_dir)

        # Stop the patchers
        self.mock_post_patcher.stop()
        self.mock_requests_post_patcher.stop()

    def test_load_from_repository(self):
        with self.assertRaises(NetorcaException):
            self.service_owner_submission.load_from_repository("/non_existent_path")

        self.service_owner_submission.load_from_repository(self.test_dir)
        self.assertIsNotNone(self.service_owner_submission.config)
        self.assertIsNotNone(self.service_owner_submission.serviceowner_submission)

    def test_load_from_repository_valid_path(self):
        self.service_owner_submission.load_from_repository(self.test_dir)
        self.assertIsNotNone(self.service_owner_submission.config)
        self.assertIsNotNone(self.service_owner_submission.serviceowner_submission)

    def test_load_from_repository_invalid_path(self):
        with self.assertRaises(NetorcaException):
            self.service_owner_submission.load_from_repository("/non_existent_path")

    def test_get_auth(self):
        with self.assertRaises(NetorcaException):
            self.service_owner_submission.get_auth()

        self.service_owner_submission.load_from_repository(self.test_dir)
        auth = self.service_owner_submission.get_auth()
        self.assertIsNotNone(auth)

    def test_get_team(self):
        with patch("netorca_sdk.auth.NetorcaAuth.get_teams_info") as mock_get_teams_info:
            mock_get_teams_info.return_value = [{"id": 1, "name": "team1"}]
            self.service_owner_submission.load_from_repository(self.test_dir)
            team = self.service_owner_submission.get_team()
            self.assertEqual(team, {"id": 1, "name": "team1"})

    def test_validate(self):
        with patch("netorca_sdk.auth.NetorcaAuth.post") as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.json.return_value = {"is_valid": True}

            with self.assertRaises(NetorcaException):
                self.service_owner_submission.validate()

            self.service_owner_submission.load_from_repository(self.test_dir)
            is_valid, errors = self.service_owner_submission.validate()
            self.assertTrue(is_valid)
            self.assertEqual(errors, "Services validated successfully.")

    def test_validate_invalid_submission(self):
        self.mock_post.return_value.status_code = 200
        self.mock_post.return_value.json.return_value = {"is_valid": False, "errors": {"key": "Invalid key"}}

        self.service_owner_submission.load_from_repository(self.test_dir)
        is_valid, errors = self.service_owner_submission.validate()
        self.assertFalse(is_valid)
        self.assertIn("Invalid services:", errors)
        self.assertIn("service1", errors)

    def test_submit(self):
        with patch("netorca_sdk.auth.NetorcaAuth.post") as mock_post:
            mock_post.return_value.status_code = 201
            with patch("requests.post") as mock_requests_post:
                mock_requests_post.return_value.status_code = 200

                mock_auth = MagicMock()
                mock_auth.fqdn = "https://example.com"
                self.service_owner_submission.auth = mock_auth

                with self.assertRaises(NetorcaException):
                    self.service_owner_submission.submit()

    def test_create_without_metadata(self):
        # Create a temporary directory for the test
        with tempfile.TemporaryDirectory() as temp_dir:
            netorca_path = os.path.join(temp_dir, ".netorca")
            os.mkdir(netorca_path)
            config_path = os.path.join(netorca_path, "config.json")

            # Create a config.json file without metadata
            config_no_metadata = {
                "netorca_global": {"base_url": "https://example.com"},
                "services": [
                    {
                        "service_name": "service1",
                        "repository_url": "https://github.com/user/repo",
                    }
                ],
            }

            with open(config_path, "w") as f:
                json.dump(config_no_metadata, f)

            # Initialize the ServiceOwnerSubmission object and load the config
            service_owner_submission = ServiceOwnerSubmission(netorca_api_key=self.netorca_api_key)
            service_owner_submission.load_from_repository(temp_dir)

            # Check if the config was loaded correctly
            self.assertIsNotNone(service_owner_submission)
            self.assertEqual(service_owner_submission.config, config_no_metadata)
            self.assertIsNotNone(service_owner_submission.auth)

    def test_create_only_readme(self):
        # Create a temporary directory for the test
        with tempfile.TemporaryDirectory() as temp_dir:
            netorca_path = os.path.join(temp_dir, ".netorca")
            os.mkdir(netorca_path)

            with open(os.path.join(temp_dir, ".netorca", "config.json"), "w") as f:
                f.write('{"netorca_global": {"base_url": "https://example.com"}}')

            with open(os.path.join(temp_dir, ".netorca", "service1.json"), "w") as f:
                f.write('{"key": "value"}')

            # Initialize the ServiceOwnerSubmission object and load the config without readme
            service_owner_submission = ServiceOwnerSubmission(netorca_api_key=self.netorca_api_key)
            service_owner_submission.load_from_repository(temp_dir)
            self.assertIsNotNone(service_owner_submission)
            self.assertIsNotNone(service_owner_submission.auth)

            # Add readme to the existing submission
            with open(os.path.join(temp_dir, ".netorca", "service1.md"), "w") as f:
                f.write("# Service 1 Documentation")
            service_owner_submission.load_from_repository(temp_dir)

            # Check if the config was loaded correctly
            self.assertIsNotNone(service_owner_submission)

    def test_pretty_print_errors(self):
        service_name = "test_service"
        errors = {
            "prop1": "Error message 1",
            "prop2": ["Error message 2", "Error message 3"],
            "nested": {
                "prop3": "Error message 4",
                "prop4": "Error message 5",
            },
        }

        with StringIO() as buf, redirect_stdout(buf):
            ServiceOwnerSubmission.pretty_print_errors(service_name, errors)
            output = buf.getvalue()

        # Check if service name and error messages are present in the output
        self.assertIn(service_name, output)
        self.assertIn("Error message 1", output)
        self.assertIn("Error message 2", output)
        self.assertIn("Error message 3", output)
        self.assertIn("Error message 4", output)
        self.assertIn("Error message 5", output)

    def test_read_json_file(self):
        # Test a non-existent file
        with self.assertRaises(NetorcaException):
            self.service_owner_submission.read_json_file("non_existent_file.json")

        # Test a valid JSON file
        valid_json_file_path = os.path.join(self.test_dir, "valid_json.json")
        with open(valid_json_file_path, "w") as f:
            f.write('{"key": "value"}')

        valid_data = self.service_owner_submission.read_json_file(valid_json_file_path)
        self.assertEqual(valid_data, {"key": "value"})

        # Test an invalid JSON file
        invalid_json_file_path = os.path.join(self.test_dir, "invalid_json.json")
        with open(invalid_json_file_path, "w") as f:
            f.write('{"key": "value",}')

        with self.assertRaises(NetorcaException):
            self.service_owner_submission.read_json_file(invalid_json_file_path)

    def test_get_service_id_by_name_valid(self):
        self.service_owner_submission.load_from_repository(self.test_dir)

        # Ensure the mock post method is targeted for get_service_id_by_name
        with patch.object(self.service_owner_submission.auth, "get") as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.json.return_value = {"results": [{"id": 123}]}

            service_id = self.service_owner_submission.get_service_id_by_name("service_name")
            self.assertEqual(service_id, 123)
            mock_get.assert_called_once_with(
                url="https://example.com/orcabase/serviceowner/services/?name=service_name",
                authentication_required=True,
            )


class TestCheckStatusCode(unittest.TestCase):
    def test_200_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": "mock_data"}
        result = ServiceOwnerSubmission.check_status_code(mock_response)
        self.assertEqual(result, {"data": "mock_data"})

    def test_204_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 204
        result = ServiceOwnerSubmission.check_status_code(mock_response)
        self.assertEqual(result, {"status": "deleted"})

    def test_404_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 404
        with self.assertRaises(NetorcaNotFoundError):
            ServiceOwnerSubmission.check_status_code(mock_response)

    def test_400_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 400
        mock_response.json.return_value = {"error": "Bad Request"}
        result = ServiceOwnerSubmission.check_status_code(mock_response)
        self.assertEqual(result, {"error": "Bad Request"})

    def test_401_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 401
        with self.assertRaises(NetorcaAuthenticationError):
            ServiceOwnerSubmission.check_status_code(mock_response)

    def test_403_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 403
        with self.assertRaises(NetorcaAuthenticationError):
            ServiceOwnerSubmission.check_status_code(mock_response)

    def test_502_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 502
        with self.assertRaises(NetorcaGatewayError):
            ServiceOwnerSubmission.check_status_code(mock_response)

    def test_503_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 503
        with self.assertRaises(NetorcaServerUnavailableError):
            ServiceOwnerSubmission.check_status_code(mock_response)

    def test_other_status_code(self):
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.content = b"Internal Server Error"
        with self.assertRaises(NetorcaAPIError):
            ServiceOwnerSubmission.check_status_code(mock_response)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch

from netorca_sdk.auth import NetorcaAuth
from netorca_sdk.exceptions import NetorcaException


class TestNetorcaAuth(unittest.TestCase):
    def setUp(self):
        self.fqdn = "https://example.com"
        self.username = "username"
        self.password = "password"
        self.api_key = "api_key"
        self.url = "https://example.com/test"
        self.data = {"key": "value"}

    @patch("netorca_sdk.auth.requests.post")
    def test_login_with_username_and_password(self, mock_post):
        mock_post.return_value.status_code = 200
        mock_post.return_value.json.return_value = {"token": "test_token"}

        auth = NetorcaAuth(self.fqdn, username=self.username, password=self.password)
        self.assertEqual(auth.token, "test_token")

    def test_login_with_api_key(self):
        auth = NetorcaAuth(self.fqdn, api_key=self.api_key)
        self.assertEqual(auth.headers["Authorization"], f"Api-Key {self.api_key}")

    @patch("netorca_sdk.auth.requests.post")
    def test_failed_authentication(self, mock_post):
        mock_post.return_value.status_code = 401
        mock_post.return_value.json.return_value = {"detail": "Invalid credentials"}

        with self.assertRaises(NetorcaException):
            NetorcaAuth(self.fqdn, username=self.username, password="wrong_password")

    @patch("netorca_sdk.auth.requests.get")
    def test_get_teams_info(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {"results": [{"id": 1, "name": "team1"}]}

        auth = NetorcaAuth(self.fqdn, api_key=self.api_key)
        teams_info = auth.get_teams_info()

        self.assertEqual(teams_info, [{"id": 1, "name": "team1"}])

    @patch("netorca_sdk.auth.requests.get")
    def test_get_teams_info_failure(self, mock_get):
        mock_get.return_value.status_code = 500
        mock_get.return_value.json.return_value = {"detail": "Internal Server Error"}

        auth = NetorcaAuth(self.fqdn, api_key=self.api_key)

        with self.assertRaises(NetorcaException):
            auth.get_teams_info()

    def test_init_missing_credentials(self):
        with self.assertRaises(NetorcaException):
            NetorcaAuth(self.fqdn)

    @patch("netorca_sdk.auth.requests.get")
    def test_get(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = {"data": "test_data"}

        auth = NetorcaAuth(self.fqdn, api_key=self.api_key)
        response = auth.get(url=self.url)

        self.assertEqual(response.json(), {"data": "test_data"})

    @patch("netorca_sdk.auth.requests.post")
    def test_post(self, mock_post):
        mock_post.return_value.status_code = 201
        mock_post.return_value.json.return_value = {"data": "test_data"}

        auth = NetorcaAuth(self.fqdn, api_key=self.api_key)
        response = auth.post(url=self.url, data=self.data)

        self.assertEqual(response.json(), {"data": "test_data"})

    @patch("netorca_sdk.auth.requests.put")
    def test_put(self, mock_put):
        mock_put.return_value.status_code = 200
        mock_put.return_value.json.return_value = {"data": "test_data"}

        auth = NetorcaAuth(self.fqdn, api_key=self.api_key)
        response = auth.put(url=self.url, data=self.data)

        self.assertEqual(response.json(), {"data": "test_data"})

    @patch("netorca_sdk.auth.requests.patch")
    def test_patch(self, mock_patch):
        mock_patch.return_value.status_code = 200
        mock_patch.return_value.json.return_value = {"data": "test_data"}

        auth = NetorcaAuth(self.fqdn, api_key=self.api_key)
        response = auth.patch(url=self.url, data=self.data)

        self.assertEqual(response.json(), {"data": "test_data"})


if __name__ == "__main__":
    unittest.main()



import json
import os
from collections import defaultdict
from string import Template
from typing import Tuple

import requests
from beautifultable import BeautifulTable

from netorca_sdk.auth import AbstractNetorcaAuth, NetorcaAuth
from netorca_sdk.config import (
    SUBMIT_SERVICEOWNER_SUBMISSION_DOCS_ENDPOINT,
    SUBMIT_SERVICEOWNER_SUBMISSION_ENDPOINT,
    VALIDATE_SERVICEOWNER_SUBMISSION_ENDPOINT,
)
from netorca_sdk.exceptions import (
    NetorcaAPIError,
    NetorcaAuthenticationError,
    NetorcaException,
    NetorcaGatewayError,
    NetorcaNotFoundError,
    NetorcaServerUnavailableError,
)


class ServiceOwnerSubmission:
    def __init__(self, netorca_api_key: str):
        self.netorca_api_key = netorca_api_key

        self.config = None
        self.serviceowner_submission = None
        self.auth = None

    def load_from_repository(
        self, repository_path: str, netorca_directory: str = ".netorca", repository_config: dict = None
    ) -> None:
        """
        Check if valid and load submission and config from service owner's repository.
        Repository must contain `.netorca` directory and `config.json` file.
        Note: Two allowed extensions in `.netorca` directory are: `*.yaml` and `*.md`.

        Args:
            repository_path: str    path to service owner repository

        Returns: None
        :param repository_path:     path to service owner repository
        :param netorca_directory:   netorca directory name, defaults to ".netorca"
        :param repository_config:      optional: you can specify config from dictionary instead of config.json
        """

        repository_exists = os.path.isdir(repository_path)
        if not repository_exists:
            raise NetorcaException(f"`{repository_path}` directory does not exist.")

        netorca_path = os.path.join(repository_path, netorca_directory)
        netorca_exists = os.path.isdir(netorca_path)
        if not netorca_exists:
            raise NetorcaException(f"`{netorca_directory}` directory does not exist.")

        if repository_config:
            netorca_global = repository_config.get("netorca_global", {})
            if not (netorca_global and netorca_global.get("base_url")):
                raise NetorcaException("No `netorca_global.base_url` provided.")
            self.config = repository_config
            self.auth = self.get_auth()
        else:
            config_path = os.path.join(netorca_path, "config.json")
            self.config = self.read_json_file(config_path)
            netorca_global = self.config.get("netorca_global", {})
            if not (netorca_global and netorca_global.get("base_url")):
                raise NetorcaException("No `netorca_global.base_url` provided.")
            self.auth = self.get_auth()

        _tmp_serviceowner_submission = defaultdict()
        for filename in os.listdir(netorca_path):
            if filename == "config.json":
                continue

            file_path = os.path.join(netorca_path, filename)
            if not os.path.isfile(file_path) and not (file_path.endswith(".json") or file_path.endswith(".md")):
                raise NetorcaException(
                    f"ServiceOwner submission file: `{file_path}` does not exist or must be `.json` or `.md` extension"
                )

            if filename.endswith(".json"):
                json_file = self.read_json_file(file_path)
                filename_without_ext = os.path.splitext(filename)[0]
                _tmp_serviceowner_submission.setdefault(filename_without_ext, {})
                _tmp_serviceowner_submission[filename_without_ext]["service"] = json_file

        for filename in os.listdir(netorca_path):
            if filename == "config.json" or not filename.endswith(".md"):
                continue

            file_path = os.path.join(netorca_path, filename)
            filename_without_ext = os.path.splitext(filename)[0]

            if not _tmp_serviceowner_submission.get(filename_without_ext):
                raise NetorcaException(
                    f"'\nYou are trying to add README file: `{filename}` for non existing service.\n"
                    f"Readme file (.md) must have the same name as service definition file (.json)."
                )

            _tmp_serviceowner_submission[filename_without_ext]["readme"] = file_path

        self.serviceowner_submission = _tmp_serviceowner_submission

    def read_json_file(self, file_path: str):
        try:
            with open(file_path, "r") as stream:
                return json.load(stream)
        except FileNotFoundError as exc:
            raise NetorcaException(f"File not found: `{file_path}`. Exception: {exc}")
        except PermissionError as exc:
            raise NetorcaException(f"Permission error while reading file: `{file_path}`. Exception: {exc}")
        except json.JSONDecodeError as exc:
            raise NetorcaException(f"Error while parsing file: `{file_path}`. Exception: {exc.msg}")

    def get_auth(self) -> AbstractNetorcaAuth:
        if not self.config:
            raise NetorcaException("Cannot authenticate before loading repository config.")

        netorca_fqdn = self.config.get("netorca_global", {}).get("base_url")
        self.auth = NetorcaAuth(fqdn=netorca_fqdn, api_key=self.netorca_api_key)
        return self.auth

    def get_team(self) -> dict:
        teams = self.auth.get_teams_info()
        if teams:
            return teams[0]
        return {}

    def validate(self, pretty_print=False) -> Tuple[bool, str]:
        """
        Validate service owner submission.
        NOTE: Data must be first imported with `load_from_repository` method

        Returns:
            bool    ->  validation successful
        """
        if self.auth is None:
            raise NetorcaException("Use `load_from_repository(repository_path)` method to load configuration.")
        VALIDATE_SERVICEOWNER_PATH = f"{self.auth.fqdn}{VALIDATE_SERVICEOWNER_SUBMISSION_ENDPOINT}"
        invalid_services = []

        if not (self.config and self.serviceowner_submission and self.auth):
            raise NetorcaException("Use `load_from_repository(repository_path)` method to load configuration.")

        for service, service_submission in self.serviceowner_submission.items():
            response = self.auth.post(
                url=VALIDATE_SERVICEOWNER_PATH,
                data=json.dumps(service_submission["service"]),
                authentication_required=True,
            )
            response = self.check_status_code(response)
            if response.get("is_valid"):
                print(f"VALIDATION SUCCESSFUL for service: `{service}`.")
            else:
                invalid_services.append(service)
                errors = response.get("errors")
                if pretty_print and errors:
                    ServiceOwnerSubmission.pretty_print_errors(service, errors)

        if invalid_services:
            print("INVALID SERVICES: " + ", ".join(invalid_services))
            return False, "Invalid services: " + ", ".join(invalid_services)
        return True, "Services validated successfully."

    def submit(self) -> Tuple[bool, str]:
        """
        Validate and submit consumer submission.
        NOTE: Data must be first imported with `load_from_repository` method

        Returns:
            bool, str    ->  is submission successful, submission messages
        """
        SUBMIT_SERVICEOWNER_REQUEST_PATH = f"{self.auth.fqdn}{SUBMIT_SERVICEOWNER_SUBMISSION_ENDPOINT}"
        submitted_services = []
        submitted_readme = []
        is_valid = self.validate(pretty_print=True)

        if not is_valid[0]:
            return False, "Some of your submissions are invalid and were not submitted."

        print()
        for service, service_submission in self.serviceowner_submission.items():
            response = self.auth.post(
                url=SUBMIT_SERVICEOWNER_REQUEST_PATH,
                data=json.dumps(service_submission["service"]),
                authentication_required=True,
            )

            if response.status_code == 201:
                print(f"SUBMITTED service to NetOrca. Service: `{service}` is now available for consumers.")
                submitted_services.append(service)
            elif response.status_code == 200:
                print("No updates to service detected")
            else:
                print(f"FAILED to submit service: `{service}`. Reason: {self.check_status_code(response)}")

            ## README's should always be submitted, regardless of whether the Service has been updated
            # The README could have changed.
            response = self.check_status_code(response)
            if service_submission.get("readme"):
                service_id = response["id"] if "id" in response else self.get_service_id_by_name(service)
                docs_path = Template(SUBMIT_SERVICEOWNER_SUBMISSION_DOCS_ENDPOINT).substitute(id=service_id)
                SUBMIT_SERVICEOWNER_DOCS_PATH = f"{self.auth.fqdn}{docs_path}"
                files = {"md_file": (f"{service}.md", open(service_submission.get("readme"), "rb"))}
                auth = self.auth.headers["Authorization"]
                response = requests.post(SUBMIT_SERVICEOWNER_DOCS_PATH, files=files, headers={"Authorization": auth})
                if response.status_code == 201:
                    print(f"SUBMITTED README file for service: `{service}`.")
                    submitted_readme.append(service)
                if response.status_code == 200:
                    print(f"UPDATED README file for service: `{service}`.")
                    submitted_readme.append(service)
                else:
                    response = self.check_status_code(response)
                    print(f"FAILED to submit README file for service: `{service}`. Reason: `{response}.`")
            else:
                print(f"NOTE: Service `{service}` does not have a README file. Skipped.")
            print("\nMoving to next service...")
            print()

        if submitted_services or submitted_readme:
            return True, "Submitted services: " + ", ".join(
                submitted_services
            ) + "\n" + "Submitted README's: " + ", ".join(submitted_readme)
        return False, "No services were submitted."

    @staticmethod
    def pretty_print_errors(service_name: str, errors: dict) -> None:
        def append_to_table(service_name: str, item: str, value: str, table: BeautifulTable) -> None:
            if isinstance(value, str) or isinstance(value, list):
                table.rows.append([service_name, item, value])
            elif isinstance(value, dict):
                for key, val in value.items():
                    append_to_table(service_name, key, val, table)

        table = BeautifulTable(maxwidth=100)
        table.set_style(BeautifulTable.STYLE_SEPARATED)
        table.columns.header = ["Schema", "Property", "Reason"]

        for item, value in errors.items():
            append_to_table(service_name, item, value, table)

        if table.rows:
            print("-" * 100)
            print(f"Schema: `{service_name}` validation errors")
            print("-" * 100)
            print(table)
            print()

    def get_service_id_by_name(self, service_name: str) -> int:
        """
        Get service id from NetOrca based on service name.

        Args:
            service_name: str   service name

        Returns:
            int     ->  service id
        """
        try:
            SUBMIT_SERVICEOWNER_REQUEST_PATH = (
                f"{self.auth.fqdn}{SUBMIT_SERVICEOWNER_SUBMISSION_ENDPOINT}?name={service_name}"
            )
            response = self.auth.get(
                url=SUBMIT_SERVICEOWNER_REQUEST_PATH,
                authentication_required=True,
            )
            response = self.check_status_code(response)
            service_id = response["results"][0]["id"] if response["results"] else None
            return service_id
        except NetorcaException:
            raise NetorcaException(f"Failed to get service id for service: `{service_name}`.")

    @staticmethod
    def check_status_code(response):
        if response.status_code in [200, 201, 400]:
            return response.json()
        elif response.status_code == 204:
            return {"status": "deleted"}
        elif response.status_code == 404:
            raise NetorcaNotFoundError("The endpoint not found.")
        elif response.status_code == 401:
            raise NetorcaAuthenticationError("Authentication failed.")
        elif response.status_code == 403:
            raise NetorcaAuthenticationError("Authorization failed.")
        elif response.status_code == 502:
            raise NetorcaGatewayError("Load balancer or webserver is down.")
        elif response.status_code == 503:
            raise NetorcaServerUnavailableError("Server is temporarily unavailable.")
        else:
            if response.content:
                error_message = response.content.decode("utf-8")
            else:
                error_message = "No content in the response."
            raise NetorcaAPIError(f"Error {response.status_code} - {error_message}")



import json
import os
from typing import Tuple

import git
import yaml
from beautifultable import BeautifulTable

from netorca_sdk.auth import AbstractNetorcaAuth, NetorcaAuth
from netorca_sdk.config import SUBMIT_CONSUMER_SUBMISSION_ENDPOINT, VALIDATE_CONSUMER_SUBMISSION_ENDPOINT
from netorca_sdk.exceptions import (
    NetorcaAPIError,
    NetorcaAuthenticationError,
    NetorcaException,
    NetorcaGatewayError,
    NetorcaNotFoundError,
    NetorcaServerUnavailableError,
    NetOrcaWrongYAMLFormat,
)


class ConsumerSubmission:
    def __init__(
        self,
        netorca_api_key: str,
        netorca_validate_only: bool = True,
        repository_url: str = "./",
        use_config: bool = False,
    ):
        self.netorca_api_key = netorca_api_key
        self.netorca_validate_only = netorca_validate_only
        self.repository_path = repository_url
        self.use_config = use_config
        self.config = None
        self.consumer_submission = None
        self.auth = None

    def load_from_repository(
        self, repository_path: str = None, netorca_directory: str = ".netorca", repository_config: dict = None
    ) -> None:
        """
        Check if valid and load request and config from consumer's repository.

        Note: Only one allowed extensions in netorca_directory directory is `*.yaml/*.yml`

        Args:
            repository_path: str    path to consumer repository

        Returns: None
        :param repository_path:     path to consumer repository
        :param netorca_directory:   netorca directory name, defaults to ".netorca"
        :param repository_config:      optional: you can specify config from dictionary instead of config.y(a)ml
        """

        if not repository_path:
            repository_path = self.repository_path
        repository_exists = os.path.isdir(repository_path)
        if not repository_exists:
            raise NetorcaException(f"`{repository_path}` directory does not exist.")

        netorca_exists = os.path.isdir(f"{repository_path}/{netorca_directory}")
        if not netorca_exists:
            raise NetorcaException(f"`{netorca_directory}` directory does not exist.")

        dotnetorca_path = f"{repository_path}/{netorca_directory}"
        if repository_config:
            netorca_global = repository_config.get("netorca_global", {})
            if not (netorca_global and netorca_global.get("base_url")):
                raise NetorcaException("No `netorca_global.base_url` provided.")

            # Check for empty base_url
            base_url = netorca_global.get("base_url", "") or ""
            base_url = base_url.strip()
            if not base_url:
                raise NetorcaException("`netorca_global.base_url` is empty.")

            self.config = repository_config
            self.auth = self.get_auth()
        else:
            # check and load config from file if it exists
            config_path_yaml = f"{repository_path}/{netorca_directory}/config.yaml"
            config_path_yml = f"{repository_path}/{netorca_directory}/config.yml"

            if os.path.exists(config_path_yml):
                config_path = config_path_yml
            elif os.path.exists(config_path_yaml):
                config_path = config_path_yaml
            else:
                raise NetorcaException("No config file in the repository.")

            config = self.load_yaml_file(config_path)
            netorca_global = config.get("netorca_global", {})
            if not (netorca_global and netorca_global.get("base_url")):
                raise NetorcaException("No `netorca_global.base_url` provided.")

            # Check for empty base_url
            base_url = netorca_global.get("base_url", "") or ""
            base_url = base_url.strip()

            if not base_url:
                raise NetorcaException("`netorca_global.base_url` is empty.")
            config["netorca_global"]["commit_id"] = self.get_commit_id(repository_path)
            self.config = config
            self.auth = self.get_auth()

        _tmp_consumer_submission = {}
        # check and load consumer request
        for filename in os.listdir(dotnetorca_path):
            if filename == "config.yaml" or filename == "config.yml":
                continue

            f = os.path.join(dotnetorca_path, filename)
            # checking if it is a file and is `*.yaml/*.yml`
            if not (os.path.isfile(f) and (f.endswith(".yaml") or f.endswith(".yml"))):
                continue

            app = self.load_yaml_file(f)

            if not isinstance(app, dict):
                raise NetorcaException(f"Invalid format in file: `{filename}`. The file should contain a dictionary.")

            for key in app.keys():
                if key in _tmp_consumer_submission:
                    raise NetorcaException(
                        f"Application with name `{key}` already exists in different yaml declaration."
                    )

                if app[key].get("services") is None:
                    app[key]["services"] = {}

                _tmp_consumer_submission.update(app)
        self.consumer_submission = _tmp_consumer_submission

    @staticmethod
    def get_commit_id(repo_path) -> str:
        try:
            repo = git.Repo(repo_path, search_parent_directories=True)
            latest_commit = repo.head.commit
            latest_commit_id = latest_commit.hexsha
            return latest_commit_id
        except git.exc.InvalidGitRepositoryError:
            return ""

    @staticmethod
    def load_yaml_file(file_path: str) -> dict:
        """
        Load yaml file from given path
        Args:
            file_path: str  path to yaml file

        Returns: dict
        """
        with open(file_path, "r") as stream:
            try:
                loaded_yaml = yaml.safe_load(stream)
                return loaded_yaml if loaded_yaml is not None else {}
            except yaml.YAMLError as exc:
                file_name = os.path.basename(file_path)
                raise NetOrcaWrongYAMLFormat(f"Error while parsing file: `{file_name}`. Exception: {exc.problem} {exc.problem_mark}")

    def get_auth(self) -> AbstractNetorcaAuth:
        if not self.config:
            raise NetorcaException("Cannot authenticate before loading repository config.")

        netorca_fqdn = self.config.get("netorca_global", {}).get("base_url")
        self.auth = NetorcaAuth(fqdn=netorca_fqdn, api_key=self.netorca_api_key)
        return self.auth

    def get_team(self) -> dict:
        if self.use_config:
            team_name = self.config.get("netorca_global", {}).get("metadata", {}).get("team_name")
            if not team_name:
                raise NetorcaException("`netorca_global.team_name` is empty.")
            return {"name": team_name}

        teams = self.auth.get_teams_info()
        if teams:
            return teams[0]
        return {}

    def prepare_request(self) -> dict:
        team = self.get_team()
        metadata = self.config.get("netorca_global", {}).get("metadata", {})
        if not (team and self.config and self.consumer_submission and self.auth):
            raise NetorcaException("Team, config and consumer request should be fetched at this stage.")

        full_request = {team["name"]: self.consumer_submission}

        if metadata is not None:
            full_request[team["name"]]["metadata"] = metadata

        return full_request if full_request else {}

    def validate(self, pretty_print=False, partial=False) -> Tuple[bool, dict]:
        """
        Validate consume request.
        NOTE: Data must be first imported with `load_from_repository` method
        Parameters:
            pretty_print:   (optional) pretty print errors, default: False
            partial:        (optional) partial validation, default: False
        Returns:
            Tuple[bool, str]    ->  is_valid, validation_errors
        """
        if not self.consumer_submission:
            print("No application detected. Validation skipped.")
            return False, {}
        if not (self.config and self.auth):
            raise NetorcaException("Use `load_from_repository(repository_path)` method to load configuration.")
        VALIDATE_REQUEST_PATH = f"{self.auth.fqdn}{VALIDATE_CONSUMER_SUBMISSION_ENDPOINT}"
        full_request = self.prepare_request()

        if partial:
            response = self.auth.patch(
                url=VALIDATE_REQUEST_PATH, data=json.dumps(full_request), authentication_required=True
            )
        else:
            response = self.auth.post(
                url=VALIDATE_REQUEST_PATH, data=json.dumps(full_request), authentication_required=True
            )

        response = self.check_status_code(response)
        if response.get("is_valid"):
            return True, {}
        errors = response.get("errors")

        if pretty_print:
            ConsumerSubmission.pretty_print_errors(errors)
        return False, errors

    def submit(self, partial=False) -> Tuple[bool, str]:
        """
        Validate and submit consumer request.
        NOTE: Data must be first imported with `load_from_repository` method

        Parameters:
            partial:        (optional) partial submission, default: False
        Returns:
            bool, str    ->  submission successful, submission messages
        """
        if not self.consumer_submission:
            print("No application detected. Submission skipped.")
            return False, "No application detected. Submission skipped."
        is_valid = self.validate(pretty_print=True, partial=partial)
        if not is_valid[0]:
            return False, "Consumer request is invalid and cannot be submitted."

        SUBMIT_REQUEST_PATH = f"{self.auth.fqdn}{SUBMIT_CONSUMER_SUBMISSION_ENDPOINT}?commit_id={self.config.get('netorca_global', {}).get('commit_id')}"
        full_request = self.prepare_request()
        if partial:
            response = self.auth.patch(
                url=SUBMIT_REQUEST_PATH, data=json.dumps(full_request), authentication_required=True
            )
        else:
            response = self.auth.post(
                url=SUBMIT_REQUEST_PATH, data=json.dumps(full_request), authentication_required=True
            )

        if response.status_code == 201:
            return True, "Submitted successfully."
        if response.status_code == 200:
            return True, "Submitted successfully, no changes detected"
        return False, self.check_status_code(response)

    def get_response(self):
        self.load_from_repository()
        return self.submit() if not self.netorca_validate_only else self.validate(pretty_print=True)

    @staticmethod
    def pretty_print_errors(errors: dict) -> None:
        """
        Pretty print errors
        #TODO: this should be refactored to cleaner code (probably recursive)
        """

        table = BeautifulTable(maxwidth=100)
        table.set_style(BeautifulTable.STYLE_SEPARATED)
        table.columns.header = ["Team", "Field", "Reason"]
        for item1, value1 in errors.items():
            if isinstance(value1, str) or isinstance(value1, list):
                table.rows.append([item1, "", value1])
            elif isinstance(value1, dict):
                for item2, value2 in value1.items():
                    if isinstance(value2, str) or isinstance(value2, list):
                        table.rows.append([item1, item2, value2])

                        if table.rows:
                            print("-" * 100)
                            print(f"Team: {item1} validation errors")
                            print("-" * 100)
                            print(table)
                            print()
                        break

        for item1, value1 in errors.items():
            if isinstance(value1, dict):
                for item2, value2 in value1.items():
                    table = BeautifulTable(maxwidth=100)
                    table.set_style(BeautifulTable.STYLE_SEPARATED)
                    table.columns.header = ["Application", "Service", "ServiceItem", "Field", "Reason"]

                    if isinstance(value2, dict):
                        for item3, value3 in value2.items():
                            if isinstance(value3, str):
                                table.rows.append([item2, "", "", item3, value3])
                            elif isinstance(value3, list):
                                for err in value3:
                                    table.rows.append([item2, "", "", item3, err])
                            elif isinstance(value3, dict):
                                for item4, value4 in value3.items():
                                    if isinstance(value4, str) or isinstance(value4, list):
                                        table.rows.append([item2, item3, "", item4, value4])
                                    elif isinstance(value4, dict):
                                        for item5, value5 in value4.items():
                                            if isinstance(value5, str) or isinstance(value5, list):
                                                table.rows.append([item2, item3, item4, item5, value5])

                        if table.rows:
                            print("-" * 100)
                            print(f"Application: `{item2}` validation errors")
                            print("-" * 100)
                            print(table)
                            print()

    @staticmethod
    def check_status_code(response):
        if response.status_code in [200, 201, 400]:
            return response.json()
        elif response.status_code == 204:
            return {"status": "deleted"}
        elif response.status_code == 404:
            raise NetorcaNotFoundError("The endpoint not found.")
        elif response.status_code == 401:
            raise NetorcaAuthenticationError("Authentication failed.")
        elif response.status_code == 403:
            raise NetorcaAuthenticationError("Authorization failed.")
        elif response.status_code == 502:
            raise NetorcaGatewayError("Load balancer or webserver is down.")
        elif response.status_code == 503:
            raise NetorcaServerUnavailableError("Server is temporarily unavailable.")
        else:
            if response.content:
                error_message = response.content.decode("utf-8")
            else:
                error_message = "No content in the response."
            raise NetorcaAPIError(f"Error {response.status_code} - {error_message}")



from abc import ABC, abstractmethod
from time import sleep

import requests
import urllib3

from netorca_sdk.config import API_VERSION, AUTH_ENDPOINT, RETRY_TIMES, TEAM_ENDPOINT
from netorca_sdk.exceptions import (
    NetorcaAPIError,
    NetorcaAuthenticationError,
    NetorcaException,
    NetorcaGatewayError,
    NetorcaNotFoundError,
    NetorcaServerUnavailableError,
)

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class AbstractNetorcaAuth(ABC):
    @abstractmethod
    def get(self, *args, **kwargs):
        ...

    @abstractmethod
    def post(self, *args, **kwargs):
        ...

    @abstractmethod
    def put(self, *args, **kwargs):
        ...

    @abstractmethod
    def patch(self, *args, **kwargs):
        ...

    @abstractmethod
    def delete(self, *args, **kwargs):
        ...

    @abstractmethod
    def options(self, *args, **kwargs):
        ...


class NetorcaAuth(AbstractNetorcaAuth):
    def __init__(self, fqdn, username=None, password=None, api_key=None):
        self.username = username
        self.password = password
        self.api_key = api_key
        self.fqdn = fqdn
        self.headers = {"content-type": "application/json"}

        if self.username and self.password:
            self.token = self.get_auth_token()
            self.headers["Authorization"] = f"Token {self.token}"
        elif self.api_key:
            self.headers["Authorization"] = f"Api-Key {self.api_key}"
        else:
            raise NetorcaException("Failed to authenticate. You must provide either (username and password) or API KEY")

    def get(self, *args, **kwargs):
        url = kwargs.get("url")
        filters = kwargs.get("filters")
        if filters:
            url = f"{url}?{'&'.join([f'{k}={v}' for k, v in filters.items()])}"
        authentication_required = kwargs.get("authentication_required", False)

        if not url:
            raise NetorcaException("URL not provided!")

        for retry in range(RETRY_TIMES):
            if authentication_required:
                response = requests.get(url=url, headers=self.headers, verify=False)
            else:
                response = requests.get(url=url, verify=False)
            if 200 <= response.status_code < 500:
                return response
            sleep(retry)
        raise NetorcaException(
            f"Failed to send GET request. GET details: {args, kwargs}. Response status_code: {response.status_code}"
        )

    def post(self, *args, **kwargs):
        url = kwargs.get("url")
        data = kwargs.get("data")
        authentication_required = kwargs.get("authentication_required", False)

        if not url or not data:
            raise NetorcaException("URL or data not provided!")

        if authentication_required:
            response = requests.post(url=url, data=data, headers=self.headers, verify=False)
        else:
            response = requests.post(url=url, data=data, verify=False)

        if 200 <= response.status_code < 500:
            return response
        raise NetorcaException(
            f"Failed to send POST request. POST details: {args, kwargs}.  Response status_code: {response.status_code}"
        )

    def put(self, *args, **kwargs):
        url = kwargs.get("url")
        data = kwargs.get("data")
        authentication_required = kwargs.get("authentication_required", False)

        if not url or not data:
            raise NetorcaException("URL or data not provided!")

        if authentication_required:
            response = requests.put(url=url, data=data, headers=self.headers, verify=False)
        else:
            response = requests.put(url=url, data=data, verify=False)
        if 200 <= response.status_code < 500:
            return response
        raise NetorcaException(
            f"Failed to send PUT request. PUT details: {args, kwargs}.  Response status_code: {response.status_code}"
        )

    def patch(self, *args, **kwargs):
        url = kwargs.get("url")
        data = kwargs.get("data")
        authentication_required = kwargs.get("authentication_required", False)

        if not url or not data:
            raise NetorcaException("URL or data not provided!")

        if authentication_required:
            response = requests.patch(url=url, data=data, headers=self.headers, verify=False)
        else:
            response = requests.patch(url=url, data=data, verify=False)

        if 200 <= response.status_code < 500:
            return response
        raise NetorcaException(
            f"Failed to send PATCH request. PATCH details: {args, kwargs}.  Response status_code: {response.status_code}"
        )

    def delete(self, *args, **kwargs):
        NetorcaException("Not implemented.")

    def options(self, *args, **kwargs):
        url = kwargs.get("url")
        authentication_required = kwargs.get("authentication_required", False)

        if not url:
            raise NetorcaException("URL not provided!")

        for retry in range(RETRY_TIMES):
            if authentication_required:
                response = requests.options(url=url, headers=self.headers, verify=False)
            else:
                response = requests.options(url=url, verify=False)
            if 200 <= response.status_code < 500:
                return response
            sleep(retry)
        raise NetorcaException(
            f"Failed to send OPTIONS request. OPTIONS details: {args, kwargs}.  Response status_code: {response.status_code}"
        )

    def get_auth_token(self) -> str:
        AUTH_URL = f"{self.fqdn}{API_VERSION}{AUTH_ENDPOINT}"

        data = {"username": self.username, "password": self.password}
        response = self.post(url=AUTH_URL, data=data, verify=False)
        if response.status_code == 200:
            return response.json()["token"]
        raise NetorcaException(f"Authentication failed due response status_code: {response.status_code}")

    def refresh_auth_token(self):
        return self.get_auth_token()

    def get_teams_info(self) -> list:
        """Get team info for given user"""
        TEAM_URL = f"{self.fqdn}{TEAM_ENDPOINT}"
        response = self.get(url=TEAM_URL, authentication_required=True)
        if response.status_code == 200:
            return response.json()["results"]
        elif response.status_code == 403:
            raise NetorcaAPIError("Access denied.")
        elif response.status_code == 404:
            raise NetorcaNotFoundError("endpoint not found.")
        elif response.status_code == 401:
            raise NetorcaAuthenticationError("Authentication failed.")
        elif response.status_code == 502:
            raise NetorcaGatewayError("Load balancer or webserver is down.")
        elif response.status_code == 503:
            raise NetorcaServerUnavailableError("Server is temporarily unavailable.")
        else:
            raise NetorcaAPIError(f"Error {response.status_code}")

    def __str__(self):
        return f"Username: {self.username}, netorca instance: {self.fqdn}."

    def __repr__(self):
        return f"Username: {self.username}, netorca instance: {self.fqdn}."
